@@ -1,9 +1,27 @@
 # Operations on Multiplyr objects
 
-if(getRversion() >= "2.15.1") {
-    # Avoid NOTEs during check about lack of global variable bindings
-    utils::globalVariables(c(".Gbase", ".end", ".expr", ".grouped", ".groups",
-        ".local", ".master", ".rows", ".start", ".tg"))
+#' Add a new column with row names
+#'
+#' @family column manipulations
+#' @param .self Data frame
+#' @param var Optional name of column
+#' @return Data frame
+#' @export
+#' @examples
+#' \donttest{
+#' dat <- Multiplyr (x=rnorm(100), alloc=1, cl=2)
+#' dat %>% add_rownames()
+#' dat %>% shutdown()
+#' }
+add_rownames <- function (.self, var="rowname") {
+    if (!is(.self, "Multiplyr")) {
+        stop ("add_rownames operation only valid for Multiplyr objects")
+    }
+
+    col <- .self$alloc_col (var)
+    .self$bm.master[, col] <- 1:nrow(.self$bm.master)
+
+    return (.self)
 }
 
 #' @rdname arrange
@@ -14,16 +32,33 @@ arrange_ <- function (.self, ..., .dots) {
     if (!is(.self, "Multiplyr")) {
         stop ("arrange operation only valid for Multiplyr objects")
     }
-    .dots <- lazyeval::all_dots (.dots, ..., all_named=TRUE)
+    .dots <- dotscombine (.dots, ...)
     if (length(.dots) == 0 || .self$empty) {
         return (.self)
     }
-    nm <- .dots2names(.dots)
-    cols <- match(nm, .self$col.names)
-    if (any(is.na(cols))) {
-        stop (sprintf("Undefined columns: %s", paste0(nm[is.na(cols)], collapse=", ")))
+
+    sort.names <- c()
+    sort.desc <- c()
+    for (i in 1:length(.dots)) {
+        if (length(.dots[[i]]) == 1) {
+            sort.names <- c(sort.names, as.character(.dots[[i]]))
+            sort.desc <- c(sort.desc, FALSE)
+        } else {
+            fn <- as.character(.dots[[i]][[1]])
+            if (length(.dots[[i]]) == 2 && fn == "desc") {
+                sort.names <- c(sort.names, as.character(.dots[[i]][[2]]))
+                sort.desc <- c(sort.desc, TRUE)
+            } else {
+                stop (paste0 ("arrange can't handle sorting expression: ", deparse(.dots[[i]])))
+            }
+        }
     }
-    .self$sort(decreasing=FALSE, .dots)
+
+    cols <- match(sort.names, .self$col.names)
+    if (any(is.na(cols))) {
+        stop (sprintf("Undefined columns: %s", paste0(sort.names[is.na(cols)], collapse=", ")))
+    }
+    .self$sort(decreasing=sort.desc, cols=cols)
     return (.self)
 }
 
@@ -33,22 +68,19 @@ define_ <- function (.self, ..., .dots) {
     if (!is(.self, "Multiplyr")) {
         stop ("arrange operation only valid for Multiplyr objects")
     }
-    .dots <- lazyeval::all_dots (.dots, ..., all_named=TRUE)
+    .dots <- dotscombine (.dots, ...)
     if (length(.dots) == 0) {
         stop ("No column names specified")
     }
     nm <- names(.dots)
-    vars <- .dots2names (.dots)
-    if (any(vars %in% .self$col.names)) {
-        stop (sprintf("Columns already defined: %s", paste0(.self$col.names[.self$col.names %in% vars], collapse=", ")))
+    if (any(nm %in% .self$col.names)) {
+        stop (sprintf("Columns already defined: %s", paste0(.self$col.names[.self$col.names %in% nm], collapse=", ")))
     }
     for (i in 1:length(.dots)) {
-        if (nm[i] == "") { #x
-            .self$alloc_col (vars[i])
-        } else {           #x=y
-            col <- .self$alloc_col (vars[i])
+        col <- .self$alloc_col (nm[i])
+        if (nm[i] != as.character(.dots[[i]])) { #x=y
             #Set type based on template
-            tpl <- as.character(.dots[[i]]$expr)
+            tpl <- as.character(.dots[[i]])
             tcol <- match (tpl, .self$col.names)
             .self$type.cols[col] <- .self$type.cols[tcol]
 
@@ -72,7 +104,7 @@ distinct_ <- function (.self, ..., .dots, auto_compact = NULL) {
         stop ("distinct operation only valid for Multiplyr objects")
     }
     if (.self$empty) { return() }
-    .dots <- lazyeval::all_dots (.dots, ..., all_named=TRUE)
+    .dots <- dotscombine (.dots, ...)
 
     if (is.null(auto_compact)) {
         auto_compact <- .self$auto_compact
@@ -81,7 +113,7 @@ distinct_ <- function (.self, ..., .dots, auto_compact = NULL) {
     N <- length (.self$cls)
 
     if (length(.dots) > 0) {
-        namelist <- .dots2names (.dots)
+        namelist <- names (.dots)
         .cols <- match(namelist, .self$col.names)
         if (any(is.na(.cols))) {
             stop (sprintf("Undefined columns: %s", paste0(namelist[is.na(.cols)], collapse=", ")))
@@ -110,8 +142,8 @@ distinct_ <- function (.self, ..., .dots, auto_compact = NULL) {
                 all(.self$bm[1, .cols] == .self$bm[2, .cols]), 0, 1)
             return (.self)
         }
-        sm1 <- bigmemory::sub.big.matrix (.self$desc, firstRow=1, lastRow=nrow(.self$bm)-1)
-        sm2 <- bigmemory::sub.big.matrix (.self$desc, firstRow=2, lastRow=nrow(.self$bm))
+        sm1 <- bigmemory.sri::attach.resource(sm_desc_comp (.self, 1))
+        sm2 <- bigmemory.sri::attach.resource(sm_desc_comp (.self, 2))
         if (length(.cols) == 1) {
             breaks <- which (sm1[,.cols] != sm2[,.cols])
         } else {
@@ -137,27 +169,48 @@ distinct_ <- function (.self, ..., .dots, auto_compact = NULL) {
 
     # (1) determine local distinct rows
     trans <- .self$cluster_eval ({
-        if (.local$empty) { return (NA) }
-        if (nrow(.local$bm) == 1) {
-            .breaks <- 1
-            return (1)
-        } else if (nrow(.local$bm) == 2) {
-            i <- ifelse (all(.local$bm[1, .cols] ==
-                                 .local$bm[2, .cols]), 1, 2)
-            .breaks <- 1:i
-            return (i)
-        }
-        .sm1 <- bigmemory::sub.big.matrix (.local$desc, firstRow=1, lastRow=nrow(.local$bm)-1)
-        .sm2 <- bigmemory::sub.big.matrix (.local$desc, firstRow=2, lastRow=nrow(.local$bm))
-        if (length(.cols) == 1) {
-            .breaks <- which (.sm1[,.cols] != .sm2[,.cols])
+        if (.local$empty) {
+            .res <- NA
         } else {
-            .breaks <- which (!apply (.sm1[,.cols] == .sm2[,.cols], 1, all))
-        }
-        rm (.sm1, .sm2)
-        .breaks <- .breaks + 1
+            if (nrow(.local$bm) == 1) {
+                .breaks <- 1
+                .res <- 1
+            } else if (nrow(.local$bm) == 2) {
+                i <- ifelse (all(.local$bm[1, .cols] ==
+                                     .local$bm[2, .cols]), 1, 2)
+                .breaks <- 1:i
+                .res <- i
+            } else {
+                .sm1 <- bigmemory.sri::attach.resource(sm_desc_comp (.local, 1))
+                .sm2 <- bigmemory.sri::attach.resource(sm_desc_comp (.local, 2))
 
-        .local$last
+                if (length(.cols) == 1) {
+                    .breaks <- which (.sm1[,.cols] != .sm2[,.cols])
+                    .breaks <- .breaks + 1
+                    .res <- .local$last
+                } else {
+                    if (nrow(.local$bm) == 1) {
+                        .breaks <- 1
+                        .res <- 1
+                    } else if (nrow(.local$bm) == 2) {
+                        i <- ifelse (all(.local$bm[1, .cols] ==
+                                             .local$bm[2, .cols]), 1, 2)
+                        .breaks <- 1:i
+                        .res <- i
+                    } else {
+                        if (length(.cols) == 1) {
+                            .breaks <- which (.sm1[,.cols] != .sm2[,.cols])
+                        } else {
+                            .breaks <- which (!apply (.sm1[,.cols] == .sm2[,.cols], 1, all))
+                        }
+                        rm (.sm1, .sm2)
+                        .breaks <- .breaks + 1
+                        .res <- .local$last
+                    }
+                }
+            }
+        }
+        .res
     })
 
     # (2) work out if there's a group change between local[1] and local[2] etc.
@@ -169,17 +222,19 @@ distinct_ <- function (.self, ..., .dots, auto_compact = NULL) {
     # (3) set breaks=1 for all where there's a transition
     .self$cluster_export_each ("tg", ".tg")
     .self$cluster_eval ({
-        if (.local$empty) { return (NULL) }
-        if (.tg) {
-            .breaks <- c(1, .breaks)
+        if (!.local$empty) {
+            if (.tg) {
+                .breaks <- c(1, .breaks)
+            }
         }
         NULL
     })
 
     # (4) filter at breaks
     .self$cluster_eval ({
-        if (.local$empty) { return (NULL) }
-        .local$filter_rows (.breaks)
+        if (!.local$empty) {
+            .local$filter_rows (.breaks)
+        }
         NULL
     })
     .self$filtered <- TRUE
@@ -219,7 +274,7 @@ filter_ <- function (.self, ..., .dots, auto_compact = NULL) {
         return (.self)
     }
 
-    .dots <- lazyeval::all_dots (.dots, ..., all_named=TRUE)
+    .dots <- dotscombine (.dots, ...)
     if (length(.dots) == 0) {
         stop ("No filtering criteria specified")
     }
@@ -230,19 +285,23 @@ filter_ <- function (.self, ..., .dots, auto_compact = NULL) {
 
     .self$cluster_export (c(".dots"))
     .self$cluster_eval ({
-        if (.local$empty) { return (NULL) }
-        if (.local$grouped) {
-            for (.g in 1:length(.local$group)) {
+        if (!.local$empty) {
+            if (.local$grouped) {
+                for (.g in .local$group) {
+                    if (.local$group_cache[.g, 3] > 0) {
+                        for (.i in 1:length(.dots)) {
+                            .local$group_restrict (.g)
+                            .res <- dotseval (.dots[.i], .local$envir())
+                            .local$filter_vector (.res[[1]])
+                            .local$group_restrict ()
+                        }
+                    }
+                }
+            } else {
                 for (.i in 1:length(.dots)) {
-                    if (.grouped[[.g]]$empty) { next }
-                    .res <- lazyeval::lazy_eval (.dots[.i], .grouped[[.g]]$envir())
-                    .grouped[[.g]]$filter_vector (.res[[1]])
+                    .res <- dotseval (.dots[.i], .local$envir())
+                    .local$filter_vector (.res[[1]])
                 }
-            }
-        } else {
-            for (.i in 1:length(.dots)) {
-                .res <- lazyeval::lazy_eval (.dots[.i], .local$envir())
-                .local$filter_vector (.res[[1]])
             }
         }
         NULL
@@ -267,8 +326,8 @@ group_by_ <- function (.self, ..., .dots, .cols=NULL, auto_partition=NULL) {
     if (.self$empty) { return (.self) }
 
     if (is.null(.cols)) {
-        .dots <- lazyeval::all_dots (.dots, ..., all_named=TRUE)
-        namelist <- .dots2names (.dots)
+        .dots <- dotscombine (.dots, ...)
+        namelist <- names (.dots)
 
         .cols <- match(namelist, .self$col.names)
         if (any(is.na(.cols))) {
@@ -293,120 +352,82 @@ group_by_ <- function (.self, ..., .dots, .cols=NULL, auto_partition=NULL) {
 
     if (nrow(.self$bm) == 1) {
         .self$bm[, .self$groupcol] <- 1
-        .self$group_sizes <- 1
+        .self$group_cache <- bigmemory::big.matrix (nrow=1, ncol=3)
+        .self$group_cache[1, ] <- c(1, 1, 1)
         .self$group_max <- 1
         return (.self)
     }
 
-    if (N == 1) {
-        if (nrow(.self$bm) == 2) {
-            i <- ifelse (all(.self$bm[1, .cols] ==
-                                 .self$bm[2, .cols]), 1, 2)
-            .self$bm[, .self$groupcol] <- 1:i
-            .self$group_sizes <- rep(1, i)
-            .self$group_max <- i
-            return (.self)
-        }
-        sm1 <- bigmemory::sub.big.matrix (.self$desc, firstRow=1, lastRow=nrow(.self$bm)-1)
-        sm2 <- bigmemory::sub.big.matrix (.self$desc, firstRow=2, lastRow=nrow(.self$bm))
-        if (length(.cols) == 1) {
-            breaks <- which (sm1[,.cols] != sm2[,.cols])
-        } else {
-            breaks <- which (!apply (sm1[,.cols] == sm2[,.cols], 1, all))
-        }
-        sizes <- c(breaks, (.self$last - .self$first)+1) - c(0, breaks)
-        breaks <- c(0, breaks) + 1
-        last <- 0
-        for (i in 1:length(breaks)) {
-            .self$bm[(last+1):breaks[i], .self$groupcol] <- i
-            last <- breaks[i]
-        }
-        .self$group_sizes <- sizes
-        .self$group_max <- length(sizes)
-        return (.self)
-    }
-
-    # (0) If partitioned by group, temporarily repartition evenly
+    #(0) If partitioned by group, temporarily repartition evenly
     regroup_partition <- .self$group_partition
-    if (.self$group_partition) {
-        .self$partition_even()
-    }
 
-    # (1) determine local groupings
-    .self$cluster_export (c(".cols"))
-    trans <- .self$cluster_eval ({
-        if (.local$empty) { return (NA) }
+    #(1) Find all breaks locally, but extend each "last" by 1 to catch
+    #    transitions
+    #(2) Add (first-1) to each break
+    #(3) Return # breaks in each cluser (b.len)
+    .self$partition_even (extend=TRUE)
+    .self$cluster_export (".cols")
+    res <- .self$cluster_eval ({
         if (nrow(.local$bm) == 1) {
-            .breaks <- 1
-            return (1)
+            .breaks <- c()
         } else if (nrow(.local$bm) == 2) {
-            i <- ifelse (all(.local$bm[1, .cols] ==
-                                 .local$bm[2, .cols]), 1, 2)
-            .local$bm[, .local$groupcol] <- 1:i
-            .breaks <- 1:i
-            .local$group_sizes <- rep(1, i)
-            .local$group_max <- i
-            return (i)
-        }
-        .sm1 <- bigmemory::sub.big.matrix (.local$desc, firstRow=1, lastRow=nrow(.local$bm)-1)
-        .sm2 <- bigmemory::sub.big.matrix (.local$desc, firstRow=2, lastRow=nrow(.local$bm))
-        if (length(.cols) == 1) {
-            .breaks <- which (.sm1[,.cols] != .sm2[,.cols])
+            .breaks <- ifelse (all(.local$bm[1, .cols] ==
+                             .local$bm[2, .cols]), c(), .local$first+1)
         } else {
-            .breaks <- which (!apply (.sm1[,.cols] == .sm2[,.cols], 1, all))
+            sm1 <- bigmemory.sri::attach.resource (sm_desc_comp (.local, 1))
+            sm2 <- bigmemory.sri::attach.resource (sm_desc_comp (.local, 2))
+            if (length(.cols) == 1 || nrow(.local$bm) == 1) {
+                .breaks <- which (sm1[, .cols] != sm2[, .cols])
+            } else {
+                .breaks <- which (!apply (sm1[, .cols] == sm2[, .cols], 1, all))
+            }
+            if (.local$first == 1) {
+                .breaks <- c(0, .breaks)
+            }
+            .breaks <- .breaks + .local$first
         }
-        rm (.sm1, .sm2)
-
-        .breaks <- c(.breaks, nrow(.local$bm))
-        .prev <- 0
-        for (.i in 1:length(.breaks)) {
-            .local$bm[(.prev+1):.breaks[.i], .local$groupcol] <- .i
-            .prev <- .breaks[.i]
-        }
-
-        .local$last
+        .length <- length(.breaks)
     })
+    b.len <- do.call (c, res)
+    G.count <- sum(b.len)
 
-    # (2) work out if there's a group change between local[1] and local[2] etc.
-    trans <- do.call (c, trans)
-    trans <- trans[-length(trans)] #last row not a transition
-    tg <- test_transition (.self, .cols, trans)
+    #(4) Allocate group_cache with nrow=sum(b.len)+1
+    .self$group_cache <- bigmemory::big.matrix (nrow=G.count, ncol=3)
 
-    # (3) add group base to each local
-    Gcount <- do.call (c, .self$cluster_eval ({
-        .local$bm[nrow(.local$bm), .local$groupcol]
-    }))
-    Gcount <- Gcount[-length(Gcount)]
+    #(5) Pass offset into group_cache[, 2] to each cluster node
+    b.off <- c(0, cumsum(b.len))[-(length(b.len)+1)] + 1
+    gcdesc <- bigmemory.sri::describe (.self$group_cache)
+    .self$cluster_export_each ("b.off", ".offset")
+    .self$cluster_export ("gcdesc", ".gcdesc")
 
-    Gtr <- rep(1, length(Gcount))
-    Gtr[tg] <- 0
-    Gbase <- cumsum(c(0, Gcount-Gtr))
-    .self$cluster_export_each ("Gbase", ".Gbase")
+    #(6) Each cluster node constructs group_cache[, 2]
     .self$cluster_eval ({
-        .local$bm[, .local$groupcol] <- .local$bm[, .local$groupcol] + .Gbase
-        .local$group <- unique (.local$bm[, .local$groupcol])
+        .local$group_cache_attach (.gcdesc)
+        if (.length > 0) {
+            .local$group_cache[.offset:(.offset+.length-1), 1] <- .breaks
+            #.local$group_cache <- bigmemory.sri::attach.resource(sm_desc_subset(.gcdesc, .offset, .offset+.length-1))
+            #.local$group_cache[, 1] <- .breaks
+        }
         NULL
     })
 
-    # (4) figure out group sizes
-    res <- .self$cluster_eval (.breaks + .local$first - 1)
-    len <- sapply (res, length)
-    clen <- cumsum(len)
-    res <- do.call (c, res)
-    if (any(!tg)) {
-        res <- res[-clen[c(!tg, FALSE)]]
+    #(7) Fill in the blanks
+    if (G.count > 1) {
+        .self$group_cache[1:(G.count-1), 2] <- .self$group_cache[2:G.count, 1] - 1
     }
-    sizes <- res - (c(0, res)[-length(res)-1])
+    .self$group_cache[G.count, 2] <- .self$last
 
-    .self$group_sizes <- sizes
-    .self$group_max <- length(sizes)
-    .self$group_sizes_stale <- FALSE
+    #(8) Calculate group sizes
+    #(9) Assign group IDs
+    #FIXME: make parallel (use calc_group_sizes?)
+    .self$group_cache[, 3] <- (.self$group_cache[, 2] - .self$group_cache[, 1]) + 1
+    .self$group_max <- G.count
+    for (i in 1:G.count) {
+        .self$bm[.self$group_cache[i, 1]:.self$group_cache[i, 2], .self$groupcol] <- i
+    }
 
-    # Input      Gcount   tg      Gbase  Output
-    # 1: G=1,2   2        FALSE   0      G=1,2
-    # 2: G=1,2   2        TRUE    1      G=2,3
-    # --transition between 2->3--
-    # 3: G=1,2                           G=4,5
+    #Needed to allow $group_restrict on master node
+    .self$group <- 1:G.count
 
     if (auto_partition && !regroup_partition) {
         .self$group_partition <- TRUE
@@ -446,7 +467,31 @@ group_sizes <- function (.self) {
         stop ("group_sizes may only be used after group_by")
     }
     .self$calc_group_sizes (delay=FALSE)
-    .self$group_sizes
+    .self$group_cache[, 3]
+}
+
+#' Return number of groups
+#'
+#' @family utility functions
+#' @param .self Data frame
+#' @return Number of groups in data frame
+#' @export
+#' @examples
+#' \donttest{
+#' dat <- Multiplyr (x=1:100, G=rep(1:4, each=25), cl=2)
+#' dat %>% group_by (G)
+#' n_groups (dat)
+#' dat %>% shutdown()
+#' }
+n_groups <- function (.self) {
+    if (!is(.self, "Multiplyr")) {
+        stop ("n_groups operation only valid for Multiplyr objects")
+    }
+    if (!.self$grouped) {
+        return (0)
+    } else {
+        return (nrow(.self$group_cache))
+    }
 }
 
 #' @rdname mutate
@@ -457,7 +502,7 @@ mutate_ <- function (.self, ..., .dots) {
     }
     if (.self$empty) { return (.self) }
 
-    .dots <- lazyeval::all_dots (.dots, ..., all_named=TRUE)
+    .dots <- dotscombine (.dots, ...)
     if (length(.dots) == 0) {
         stop ("No mutation operations specified")
     }
@@ -468,24 +513,29 @@ mutate_ <- function (.self, ..., .dots) {
         if (.self$grouped) {
             stop("mutate on a group column is not permitted")
         } else {
-            .self$group.cols <- 0
+            .self$group.cols <- numeric(0)
         }
     }
 
     .self$cluster_export (c(".resnames", ".rescols", ".dots"))
     .self$cluster_eval ({
-        if (.local$empty) { return (NULL) }
-        if (.local$grouped) {
-            for (.g in 1:length(.local$group)) {
+        if (!.local$empty) {
+            if (.local$grouped) {
+                for (.g in .local$group) {
+                    if (.local$group_cache[.g, 1] > 0) {
+                        for (.i in 1:length(.dots)) {
+                            .local$group_restrict (.g)
+                            .res <- dotseval (.dots[.i], .local$envir())
+                            .local$set_data (, .rescols[.i], .res[[1]])
+                            .local$group_restrict ()
+                        }
+                    }
+                }
+            } else {
                 for (.i in 1:length(.dots)) {
-                    .res <- lazyeval::lazy_eval (.dots[.i], .grouped[[.g]]$envir())
-                    .grouped[[.g]]$set_data (, .rescols[.i], .res[[1]])
+                    .res <- dotseval (.dots[.i], .local$envir())
+                    .local$set_data (, .rescols[.i], .res[[1]])
                 }
-            }
-        } else {
-            for (.i in 1:length(.dots)) {
-                .res <- lazyeval::lazy_eval (.dots[.i], .local$envir())
-                .local$set_data (, .rescols[.i], .res[[1]])
             }
         }
         NULL
@@ -553,7 +603,7 @@ partition_group_ <- function (.self, ..., .dots) {
     if (!is(.self, "Multiplyr")) {
         stop ("partition_group operation only valid for Multiplyr objects")
     }
-    .dots <- lazyeval::all_dots (.dots, ..., all_named=TRUE)
+    .dots <- dotscombine (.dots, ...)
 
     if (length(.dots) > 0) {
         .self$group_partition <- TRUE
@@ -566,7 +616,7 @@ partition_group_ <- function (.self, ..., .dots) {
 
     N <- length(.self$cls)
 
-    G <- .self$group_sizes
+    G <- .self$group_cache[, 3]
     if (length(G) == 1) {
         Gi <- distribute (1, N)
         Gi[Gi == 0] <- NA
@@ -582,12 +632,11 @@ partition_group_ <- function (.self, ..., .dots) {
     .self$cluster_eval ({
         if (NA %in% .groups) {
             .local$empty <- TRUE
-            return (NULL)
         }
 
-        .local <- .master$copy(shallow=TRUE)
-        .local$empty <- FALSE
-        .local$group <- .groups
+        if (!.local$empty) {
+            .local$group <- .groups
+        }
 
         NULL
     })
@@ -605,7 +654,7 @@ reduce_ <- function (.self, ..., .dots, auto_compact = NULL) {
     if (!is(.self, "Multiplyr")) {
         stop ("reduce operation only valid for Multiplyr objects")
     }
-    .dots <- lazyeval::all_dots (.dots, ..., all_named=TRUE)
+    .dots <- dotscombine (.dots, ...)
     if (length(.dots) == 0) {
         stop ("No reduce operations specified")
     }
@@ -629,21 +678,22 @@ reduce_ <- function (.self, ..., .dots, auto_compact = NULL) {
     if (!.self$empty) {
         if (.self$grouped) {
             for (g in 1:.self$group_max) {
-                grp <- .self$group_restrict (g)
-                res <- lazyeval::lazy_eval (.dots, grp$envir())
+                .self$group_restrict (g)
+                res <- dotseval (.dots, .self$envir())
                 len <- 0
                 for (i in 1:length(res)) {
-                    grp$bm[, newcols[i]] <- res[[i]]
+                    .self$bm[, newcols[i]] <- res[[i]]
                     if (length(res[[i]]) > len) {
                         len <- length(res[[i]])
                     }
                 }
-                grp$bm[, grp$filtercol] <- 0
-                grp$bm[1:len, grp$filtercol] <- 1
-                grp$filtered <- TRUE
+                .self$bm[, .self$filtercol] <- 0
+                .self$bm[1:len, .self$filtercol] <- 1
+                .self$filtered <- TRUE
+                .self$group_restrict ()
             }
         } else {
-            res <- lazyeval::lazy_eval (.dots, .self$envir())
+            res <- dotseval (.dots, .self$envir())
             len <- 0
             for (i in 1:length(res)) {
                 .self$bm[, newcols[i]] <- res[[i]]
@@ -693,7 +743,7 @@ regroup <- function (.self, auto_partition=NULL) {
         warning ("regroup attempted on an object that's already grouped")
         return (.self)
     }
-    if (.self$group.cols == 0) {
+    if (length(.self$group.cols) == 0) {
         stop ("regroup may only be used after group_by (and without modifying the group columns)")
     }
 
@@ -720,12 +770,12 @@ rename_ <- function (.self, ..., .dots) {
     if (!is(.self, "Multiplyr")) {
         stop ("rename operation only valid for Multiplyr objects")
     }
-    .dots <- lazyeval::all_dots (.dots, ..., all_named=TRUE)
+    .dots <- dotscombine (.dots, ...)
     if (length(.dots) == 0) {
         stop ("No renaming operations specified")
     }
     newnames <- names(.dots)
-    oldnames <- as.vector (vapply (.dots, function (x) { as.character (x$expr) }, ""))
+    oldnames <- simplify2array (as.character(.dots))
     m <- match(oldnames, .self$col.names)
     if (any(is.na(m))) {
         stop (sprintf("Undefined columns: %s", paste0(oldnames[is.na(m)], collapse=", ")))
@@ -741,11 +791,11 @@ select_ <- function (.self, ..., .dots) {
     if (!is(.self, "Multiplyr")) {
         stop ("select operation only valid for Multiplyr objects")
     }
-    .dots <- lazyeval::all_dots (.dots, ..., all_named=TRUE)
+    .dots <- dotscombine (.dots, ...)
     if (length(.dots) == 0) {
         stop ("No select columns specified")
     }
-    coln <- as.vector (vapply (.dots, function (x) { as.character (x$expr) }, ""))
+    coln <- simplify2array (as.character(.dots))
     cols <- match (coln, .self$col.names)
     if (any(is.na(cols))) {
         stop (sprintf("Undefined columns: %s", paste0(coln[is.na(cols)], collapse=", ")))
@@ -833,39 +883,54 @@ slice <- function (.self, rows=NULL, start=NULL, end=NULL, each=FALSE, auto_comp
         if (is.null(rows)) {
             .self$cluster_export (c("start", "end"), c(".start", ".end"))
             .self$cluster_eval ({
-                if (.local$empty) { return (NULL) }
-                if (.local$grouped) {
-                    for (.g in 1:length(.local$group)) {
-                        .grouped[[.g]]$filter_range (.start, .end)
+                if (!.local$empty) {
+                    if (.local$grouped) {
+                        for (.g in .local$group) {
+                            if (.local$group_cache[.g, 1] > 0) {
+                                .local$group_restrict (.g)
+                                .local$filter_range (.start, .end)
+                                .local$group_restrict()
+                            }
+                        }
+                    } else {
+                        .local$filter_range (.start, .end)
                     }
-                } else {
-                    .local$filter_range (.start, .end)
                 }
                 NULL
             })
         } else if (is.logical(rows)) {
             .self$cluster_export ("rows", ".rows")
             .self$cluster_eval ({
-                if (.local$empty) { return (NULL) }
-                if (.local$grouped) {
-                    for (.g in 1:length(.local$group)) {
-                        .grouped[[.g]]$filter_vector (.rows)
+                if (!.local$empty) {
+                    if (.local$grouped) {
+                        for (.g in .local$group) {
+                            if (.local$group_cache[.g, 3] > 0) {
+                                .local$group_restrict (.g)
+                                .local$filter_vector (.rows)
+                                .local$group_restrict ()
+                            }
+                        }
+                    } else {
+                        .local$filter_vector (.rows)
                     }
-                } else {
-                    .local$filter_vector (.rows)
                 }
                 NULL
             })
         } else {
             .self$cluster_export ("rows", ".rows")
             .self$cluster_eval ({
-                if (.local$empty) { return (NULL) }
-                if (.local$grouped) {
-                    for (.g in 1:length(.local$group)) {
-                        .grouped[[.g]]$filter_rows (.rows)
+                if (!.local$empty) {
+                    if (.local$grouped) {
+                        for (.g in .local$group) {
+                            if (.local$group_cache[.g, 3] > 0) {
+                                .local$group_restrict (.g)
+                                .local$filter_rows (.rows)
+                                .local$group_restrict ()
+                            }
+                        }
+                    } else {
+                        .local$filter_rows (.rows)
                     }
-                } else {
-                    .local$filter_rows (.rows)
                 }
                 NULL
             })
@@ -896,7 +961,7 @@ summarise_ <- function (.self, ..., .dots, auto_compact = NULL) {
     if (!is(.self, "Multiplyr")) {
         stop ("summarise operation only valid for Multiplyr objects")
     }
-    .dots <- lazyeval::all_dots (.dots, ..., all_named=TRUE)
+    .dots <- dotscombine (.dots, ...)
     if (length(.dots) == 0) {
         stop ("No summarise operations specified")
     }
@@ -919,33 +984,38 @@ summarise_ <- function (.self, ..., .dots, auto_compact = NULL) {
 
     .self$cluster_export (c(".dots", ".newcols"))
     .self$cluster_eval ({
-        if (.local$empty) { return (NULL) }
-        if (.local$grouped) {
-            for (.g in 1:length(.grouped)) {
-                .res <- lazyeval::lazy_eval (.dots, .grouped[[.g]]$envir())
-                .len <- 0
-                for (.i in 1:length(.res)) {
-                    .grouped[[.g]]$bm[, .newcols[.i]] <- .res[[.i]]
-                    if (length(.res[[.i]]) > .len) {
-                        .len <- length(.res[[.i]])
+        if (!.local$empty) {
+            if (.local$grouped) {
+                for (.g in .local$group) {
+                    if (.local$group_cache[.g, 1] > 0) {
+                        .local$group_restrict (.g)
+                        .res <- dotseval (.dots, .local$envir())
+                        .len <- 0
+                        for (.i in 1:length(.res)) {
+                            .local$bm[, .newcols[.i]] <- .res[[.i]]
+                            if (length(.res[[.i]]) > .len) {
+                                .len <- length(.res[[.i]])
+                            }
+                        }
+                        .local$bm[, .local$filtercol] <- 0
+                        .local$bm[1:.len, .local$filtercol] <- 1
+                        .local$filtered <- TRUE
+                        .local$group_restrict()
                     }
                 }
-                .grouped[[.g]]$bm[, .grouped[[.g]]$filtercol] <- 0
-                .grouped[[.g]]$bm[1:.len, .grouped[[.g]]$filtercol] <- 1
-                .grouped[[.g]]$filtered <- TRUE
-            }
-        } else {
-            .res <- lazyeval::lazy_eval (.dots, .local$envir())
-            .len <- 0
-            for (.i in 1:length(.res)) {
-                .local$bm[, .newcols[.i]] <- .res[[.i]]
-                if (length(.res[[.i]]) > .len) {
-                    .len <- length(.res[[.i]])
+            } else {
+                .res <- dotseval (.dots, .local$envir())
+                .len <- 0
+                for (.i in 1:length(.res)) {
+                    .local$bm[, .newcols[.i]] <- .res[[.i]]
+                    if (length(.res[[.i]]) > .len) {
+                        .len <- length(.res[[.i]])
+                        .local$bm[, .local$filtercol] <- 0
+                        .local$bm[1:.len, .local$filtercol] <- 1
+                        .local$filtered <- TRUE
+                    }
                 }
             }
-            .local$bm[, .local$filtercol] <- 0
-            .local$bm[1:.len, .local$filtercol] <- 1
-            .local$filtered <- TRUE
         }
         NULL
     })
@@ -971,43 +1041,45 @@ transmute_ <- function (.self, ..., .dots) {
 
     if (.self$empty) { return (.self) }
 
-    .dots <- lazyeval::all_dots (.dots, ..., all_named=TRUE)
+    .dots <- dotscombine (.dots, ...)
     if (length(.dots) == 0) {
         stop ("No mutation operations specified")
     }
 
     #mutate
     .resnames <- names(.dots)
-    .rescols <- .self$alloc_col (.resnames, update=FALSE)
+    .rescols <- .self$alloc_col (.resnames, update=TRUE)
     if (any(.rescols == .self$group.cols)) {
         if (.self$grouped) {
             stop("transmute on a group column is not permitted")
         } else {
-            .self$group.cols <- 0 #prevent regroup
+            .self$group.cols <- numeric(0)
         }
     }
 
     .self$cluster_export (c(".resnames", ".rescols", ".dots"))
     .self$cluster_eval ({
-        if (.local$empty) { return (NULL) }
-        if (.local$grouped) {
-            for (.g in 1:length(.local$group)) {
+        if (!.local$empty) {
+            if (.local$grouped) {
+                for (.g in .local$group) {
+                    if (.local$group_cache[.g, 1] > 0) {
+                        for (.i in 1:length(.dots)) {
+                            .local$group_restrict (.g)
+                            .res <- dotseval (.dots[.i], .local$envir())
+                            .local$set_data (, .rescols[.i], .res[[1]])
+                            .local$group_restrict ()
+                        }
+                    }
+                }
+            } else {
                 for (.i in 1:length(.dots)) {
-                    .res <- lazyeval::lazy_eval (.dots[.i], .grouped[[.g]]$envir())
-                    .grouped[[.g]]$set_data (, .rescols[.i], .res[[1]])
+                    .res <- dotseval (.dots[.i], .local$envir())
+                    .local$set_data (, .rescols[.i], .res[[1]])
                 }
-            }
-        } else {
-            for (.i in 1:length(.dots)) {
-                .res <- lazyeval::lazy_eval (.dots[.i], .local$envir())
-                .local$set_data (, .rescols[.i], .res[[1]])
             }
         }
         NULL
     })
-    if (any(.rescols == .self$group.cols)) {
-        .self$calc_group_sizes()
-    }
     #/mutate
 
     dropcols <- .self$order.cols > 0
@@ -1030,12 +1102,12 @@ undefine_ <- function (.self, ..., .dots) {
         stop ("undefine operation only valid for Multiplyr objects")
     }
 
-    .dots <- lazyeval::all_dots (.dots, ..., all_named=TRUE)
+    .dots <- dotscombine (.dots, ...)
     if (length(.dots) == 0) {
         stop ("No undefine operations specified")
     }
 
-    dropnames <- .dots2names (.dots)
+    dropnames <- names (.dots)
     dropcols <- match (dropnames, .self$col.names)
     if (any(is.na(dropcols))) {
         stop (sprintf("Undefined columns: %s", paste0(dropnames[is.na(dropcols)], collapse=", ")))
@@ -1117,9 +1189,14 @@ within_group <- function (.self, expr) {
     expr <- substitute(expr)
     .self$cluster_export ("expr", ".expr")
     .self$cluster_eval({
-        if (.local$empty) { return(NULL) }
-        for (.g in 1:length(.local$group)) {
-            eval (.expr, envir = .grouped[[.g]]$envir())
+        if (!.local$empty) {
+            for (.g in .local$group) {
+                .local$group_restrict (.g)
+                if (!.local$empty) {
+                    eval (.expr, envir = .local$envir())
+                }
+                .local$group_restrict ()
+            }
         }
         NULL
     })
@@ -1145,8 +1222,9 @@ within_node <- function (.self, expr) {
     expr <- substitute(expr)
     .self$cluster_export ("expr", ".expr")
     .self$cluster_eval({
-        if (.local$empty) { return(NULL) }
-        eval (.expr, envir = .local$envir())
+        if (!.local$empty) {
+            eval (.expr, envir = .local$envir())
+        }
         NULL
     })
     .self
